
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <GL/glut.h>
#include <GL/gl.h>

#include <sys/timeb.h>

#include "cam.h"

const float DEG2RAD = 0.017453292519943;
const float RAD2DEG = 57.29577951308232;

float plasticity = 2;     // Heights scale factor
float gridX      = 0.1;   // OpenGL units per height point grid
float gridY      = 0.1;   // OpenGL units per height point grid
float angle      = 50;    // Perpective view angle
float speedXY    = 0.5;   // Camera translate speed in OpenGL units per keypress
float speedZ     = 0.25;  // Camera translate speed in OpenGL units per keypress
float rotate     = 5;     // Camera rotate speed in degrees per keypress
float swingstep  = 0.1;
float swingdelay = 200;
float swingampl  = 0.5;
float eyes       = 0.2;

Camera cam;

float swing = 0;
float translateX = 0;
float translateY = 0;
float translateZ = 0;
float slideX = 0;
float slideY = 0;
float slideZ = 0;
float rotateX = 0;
float rotateY = 0;
float rotateZ = 0;

const int textureX = 128;
const int textureY = 128;

unsigned short *heights = NULL;
unsigned char *palette = NULL;
typedef struct { float x, y; } trackpoint;
trackpoint *track = NULL;

char name_heights[256],name_surface[256],name_palette[256],name_track[256];
int surface_stopY,surface_startX,surface_startY,surface_stopX;
int heights_startX,heights_stopX,heights_startY,heights_stopY;
int heights_stepX,heights_stepY;
int heightsX,heightsY;
int overlapX,overlapY;
int actX,actY;
int pixelsX,pixelsY;
int startX,startY,stopX,stopY;

float heightfactor;
bool surface;
int tracklen;

int list_solid, list_wire, list_track, terrain;
int win_main, win_info;
float ratio;
timeb last = { 0, 0, 0, 0 };

// SETUP
bool solid = false;
bool antialias = true;
bool light = true;
bool textures = true;
bool smooth = true;
bool flow = true;
bool status = true;
bool stereo = false;
bool reflector = false;
enum { MORNING, NOON, AFTERNOON} lightstyle = AFTERNOON;

class Vector3f {
private:
  GLfloat data[3];
public:
  void Set(float x, float y, float z) {
    data[0] = x;
    data[1] = y;
    data[2] = z;
  }
  void Zero() {
    data[0] = 0;
    data[1] = 0;
    data[2] = 0;
  }
  void CrossProduct(const Vector3f &a,const Vector3f &b) {
    data[0] = a.data[1]*b.data[2]-a.data[2]*b.data[1];
    data[1] = a.data[2]*b.data[0]-a.data[0]*b.data[2];
    data[2] = a.data[0]*b.data[1]-a.data[1]*b.data[0];
  }
  Vector3f &operator+=(const Vector3f &a) {
    data[0] += a.data[0];
    data[1] += a.data[1];
    data[2] += a.data[2];
    return *this;
  }
  Vector3f &operator-=(const Vector3f &a) {
    data[0] -= a.data[0];
    data[1] -= a.data[1];
    data[2] -= a.data[2];
    return *this;
  }
  Vector3f operator-(const Vector3f &a) {
    Vector3f result;
    result.data[0] = data[0] - a.data[0];
    result.data[1] = data[1] - a.data[1];
    result.data[2] = data[2] - a.data[2];
    return result;
  }
  void Normalize() {
    float length = sqrt(data[0]*data[0]+data[1]*data[1]+data[2]*data[2]);
    data[0] /= length;
    data[1] /= length;
    data[2] /= length;
  }
  float& operator[](int i) {
    return data[i];
  }
  inline operator const float* () {
    return data;
  }
};

Vector3f *normals;

void exit();

void* LoadWholeFile(const char *name) {
  FILE* file = fopen(name,"rb");
  if (!file) {
    printf("%s not found\n",name);
    exit(0);
  }
  fseek(file,0,SEEK_END); int size = ftell(file); fseek(file,0,SEEK_SET);
  unsigned char* buffer = new unsigned char[size];
  fread(buffer,1,size,file);
  fclose(file);
  return buffer;
}

void* LoadPartOfFile(const char *name, int columns, int rows, int left, int top, int width, int height) {
  FILE* file = fopen(name,"rb");
  if (!file) { printf("%s not found\n",name); exit(0); }
  
  unsigned char* buffer = new unsigned char[width*height];
  memset(buffer,0,sizeof(unsigned char)*width*height);
  for (int y=0; y<height; y++) {
    if (top+y<rows) {
      int now = (width<columns-left) ? width : columns-left;
      fseek(file,(top+y)*columns+left,SEEK_SET);
      fread(&(buffer[y*width]),1,now,file);
    }
  }
  fclose(file);

  return buffer;
}

void DrawString(const char *s, float x, float y) {
  glRasterPos2f(x,y);
  while (*s) glutBitmapCharacter(GLUT_BITMAP_HELVETICA_12, *(s++));
};

inline int roundup(float x) {
  double dummy;
  if (modf(x,&dummy)==0) return (int)x;
  else return (int)x+1
;
}

inline int tex2heightsX(int x) {
  int pixelX = x*(textureX-overlapX);
  int geoX = surface_startX+(surface_stopX-surface_startX)*pixelX/pixelsX;
  return roundup((float)(geoX-heights_startX)/heights_stepX);
}

inline int tex2heightsY(int y) {
  int pixelY = y*(textureY-overlapY);
  int geoY = surface_startY+(surface_stopY-surface_startY)*pixelY/pixelsY;
  return roundup((float)(geoY-heights_startY)/heights_stepY);
}

inline float heights2pixelX(int x) {
  int geoX = heights_stepX*x+heights_startX;
  return (float)(pixelsX*(geoX-surface_startX))/(surface_stopX-surface_startX);
}

inline float heights2pixelY(int y) {
  int geoY = heights_stepY*y+heights_startY;
  return (float)(pixelsY*(geoY-surface_startY))/(surface_stopY-surface_startY);
}

inline int heights2texX(int x) {
  return (int)(heights2pixelX(x)/(textureX-overlapX));
}

inline int heights2texY(int y) {
  return (int)(heights2pixelY(y)/(textureY-overlapY));
}

inline float heights2acttexX(int x) {
  return (heights2pixelX(x)-actX*(textureX-overlapX))/(float)textureX;
}

inline float heights2acttexY(int y) {
  return (heights2pixelY(y)-actY*(textureY-overlapY))/(float)textureY;
}

void LoadTexture(int x, int y) {
  unsigned char* texture;
  if (palette) {
    unsigned char *indexes = (unsigned char*)LoadPartOfFile(name_surface,
      pixelsX,pixelsY,
      x*(textureX-overlapX),y*(textureY-overlapY),
      textureX,textureY);
    texture = new unsigned char[3*textureX*textureY];
    for (int i=0; i<textureX*textureY; i++) {
      int index = indexes[i];
      texture[3*i+0]=palette[3*index+0];
      texture[3*i+1]=palette[3*index+1];
      texture[3*i+2]=palette[3*index+2];
    }
  }
  else {
    texture = (unsigned char*)LoadPartOfFile(name_surface,
      3*pixelsX,pixelsY,
      x*3*(textureX-overlapX),y*(textureY-overlapY),
      3*textureX,textureY);
  }
  glTexImage2D(GL_TEXTURE_2D, 0, 3, textureX, textureY, 0, GL_RGB, GL_UNSIGNED_BYTE, texture);
  delete texture;
  actX = x; actY = y;
}

trackpoint* LoadTrack(const char* name) {
  printf("Loading track .. \n");
  float northing, easting;
  FILE* file = fopen(name,"rb");
  if (!file) { printf("%s not found\n",name); exit(); }
  tracklen = 0;
  while (!feof(file)) {
    if (fscanf(file,"%f %f\n",&northing,&easting)<2) {
      printf("format error\n");
      exit();
    }
    tracklen++;
  }
  trackpoint *track = new trackpoint[tracklen];
  fseek(file,0,SEEK_SET);
  int i = 0;
  while (!feof(file)) {
    if (fscanf(file,"%f %f\n",&(track[i].y),&(track[i].x))<2) {;
      printf("format error\n");
      exit();
    }
    i++;
  }
  fclose(file);
  return track;
}

inline bool GetPointHeight(int x, int y, float &height) {
  if ((x<0)||(x>=heightsX)||(y<0)||(y>=heightsY)) return false;
  height = heights[y*heightsX+x];
  if (height==0xFC19) height = 0;
  return true;
}

inline bool GetTerrainPoint(int x,int y,Vector3f &point) {
  float height;
  if (!GetPointHeight(x,y,height)) return false;
  point.Set(gridX*x,gridY*y,heightfactor*height);
  return true;
}

bool InterpolateHeight(float x, float y, float &height) {
  float x1 = (int)x; float x2 = x1+1;
  float y1 = (int)y; float y2 = y1+1;
  float h11,h12,h21,h22;
  if (GetPointHeight((int)x1,(int)y1,h11) && GetPointHeight((int)x1,(int)y2,h12) &&
      GetPointHeight((int)x2,(int)y1,h21) && GetPointHeight((int)x2,(int)y2,h22)) {
    float h1 = h11+(x-x1)*(h21-h11);
    float h2 = h12+(x-x1)*(h22-h12);
    height   = h1+(y-y1)*(h2-h1);
    return true;
  }
  return false;
}

bool InterpolateTerrainPoint(float x,float y,Vector3f &point) {
  float height;
  if (!InterpolateHeight(x,y,height)) return false;
  point.Set(gridX*x,gridY*y,heightfactor*height);
  return true;
}

void ComputeTerrainNormal(int x,int y,Vector3f &result) {
  result.Zero();
  Vector3f central,p1,p2,p3,p4,normal;
  GetTerrainPoint(x,y,central);
  bool is_p1 = GetTerrainPoint(x+1,y  ,p1);
  bool is_p2 = GetTerrainPoint(x  ,y+1,p2);
  bool is_p3 = GetTerrainPoint(x-1,y  ,p3);
  bool is_p4 = GetTerrainPoint(x  ,y-1,p4);
  p1 -= central; p2 -= central;
  p3 -= central; p4 -= central;
  if (is_p1 && is_p2) {
    normal.CrossProduct(p1,p2);
    result+=normal;
  }
  if (is_p2 && is_p3) {
    normal.CrossProduct(p2,p3);
    result+=normal;
  }
  if (is_p3 && is_p4) {
    normal.CrossProduct(p3,p4);
    result+=normal;
  }
  if (is_p4 && is_p1) {
    normal.CrossProduct(p4,p1);
    result+=normal;
  }
  result.Normalize();
}

inline Vector3f& GetCachedNormal(int x, int y) {
  return normals[y*heightsX+x];
}

inline void GeneratePoint(int x,int y) {
  Vector3f vector;
  if (GetTerrainPoint(x,y,vector)) {
    glTexCoord2f(heights2acttexX(x),heights2acttexY(y));
    if (((heights_stepX*x+heights_startX)%1000==0) || 
        ((heights_stepY*y+heights_startY)%1000==0)) {
      glColor3f(1.0,0.0,0.0);
    }
    else {
      glColor3f(1.0,1.0,1.0);
    }
    glNormal3fv(GetCachedNormal(x,y));
    glVertex3fv(vector);
  }
}

inline void GenerateInterpolatedPoint(float x,float y,float dz = 0) {
  Vector3f vector;
  if (InterpolateTerrainPoint(x,y,vector)) {
    glNormal3fv(GetCachedNormal((int)x,(int)y));
 // "nearest less" computed normal
    vector[2]+=dz;
    glVertex3fv(vector);
  }
}

void GenerateSolidTerrain(int startX,int startY,int stopX,int stopY) {
  for (int y=startY; y<stopY-1; y++) {
    glBegin(GL_TRIANGLE_STRIP);
    for(int x=startX; x<stopX; x++) {
      GeneratePoint(x,y);
      GeneratePoint(x,y+1);
    }
    glEnd();
  }
}

void GenerateWireTerrain(int startX,int startY,int stopX,int stopY) {
  int x,y;
  glLineWidth(1);
  for(y=startY; y<stopY; y++) {
    glBegin(GL_LINE_STRIP);
    for(x=startX; x<stopX; x++) GeneratePoint(x,y);
    glEnd();
  }
  for(x=startX; x<stopX; x++) {
    glBegin(GL_LINE_STRIP);
    for(y=startY; y<stopY; y++) GeneratePoint(x,y);
    glEnd();
  }
}

void GenerateTerrain() {
  int x,y;

  printf("Computing normals .. %d vertexes\n",heightsX*heightsY);
  for(y=0; y<heightsY; y++)
    for(x=0; x<heightsX; x++)
      ComputeTerrainNormal(x,y,normals[y*heightsX+x]);

  glNewList(list_solid = glGenLists(1),GL_COMPILE);
  if (surface) {
    int texcount = (heights2texY(stopY)-heights2texY(startY))*(heights2texX(stopX)-heights2texX(startX));
    float texsize = (float)(texcount*3*textureX*textureY)/1024/1024;
    printf("Generating solid textured terrain .. %d textures of %.1f MB total\n",texcount,texsize);
    for (y=heights2texY(startY); y<heights2texY(stopY); y++) {
      for (x=heights2texX(startX); x<heights2texX(stopX); x++) {
        LoadTexture(x,y);
        GenerateSolidTerrain(tex2heightsX(x)    ,tex2heightsY(y),
                             tex2heightsX(x+1)+1,tex2heightsY(y+1)+1);
      }
    }
  }
  else {
    printf("Generating solid terrain .. \n");
    GenerateSolidTerrain(startX,startY,stopX,stopY);
  }
  glEndList();

  glNewList(list_wire = glGenLists(1),GL_COMPILE);
  if (surface) {
    int vertexes =
      (tex2heightsX(heights2texX(stopX))-tex2heightsX(heights2texX(startX))+1)*
      (tex2heightsY(heights2texY(stopY))-tex2heightsY(heights2texY(startY))+1);
    printf("Generating wire terrain .. %d vertexes\n", vertexes);
    for (y=heights2texY(startY); y<heights2texY(stopY); y++) {
      for (x=heights2texX(startX); x<heights2texX(stopX); x++) {
        GenerateWireTerrain(tex2heightsX(x)    ,tex2heightsY(y),
                            tex2heightsX(x+1)+1,tex2heightsY(y+1)+1);
      }
    }
  }
  else {
    int vertexes = 
    printf("Generating wire terrain .. %d vertexes\n", (stopX-startX)*(stopY-startY));
    GenerateWireTerrain(startX,startY,stopX,stopY);
  }
  glEndList();
}

void GenerateTrack() {
  if (!track) return;
  printf("Generating track ..\n");

  glNewList(list_track = glGenLists(1),GL_COMPILE);
  glColorMaterial(GL_FRONT, GL_DIFFUSE);
  glEnable(GL_COLOR_MATERIAL);
  glColor3f(1,0,0);
  glLineWidth(2);
  glBegin(GL_LINE_STRIP);
  for(int i=0; i<tracklen; i++) {
    GenerateInterpolatedPoint(
      (float)(track[i].x-heights_startX)/heights_stepX,
      (float)(track[i].y-heights_startY)/heights_stepY,
      heightfactor*4);
  }
  glEnd();
  glDisable(GL_COLOR_MATERIAL);

  glEndList();
}

void Transformations() {
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(angle, ratio, 0.1, 100.0);
  glMatrixMode(GL_MODELVIEW);
  cam.Update_glMatrix();
}

void Translate(float x, float y, float z) {
  if (flow) {
    translateX += x;
    translateY += y;
    translateZ += z;
  }
  else {
    cam.Translate(x,y,z);
    glutPostRedisplay();
  }
}

void Slide(float x, float y, float z) {
  if (flow) {
    slideX += x;
    slideY += y;
    slideZ += z;
  }
  else {
    cam.Slide(x,y,z);
    glutPostRedisplay();
  }
}

void Rotate(float angle, axis axe) {
  if (flow) {
    switch (axe) {
      case X_AXIS : rotateX += angle; return;
      case Y_AXIS : rotateY += angle; return;
      case Z_AXIS : rotateZ += angle; return;
    }
  }
  else {
    cam.Rotate(angle, axe);
    glutPostRedisplay();
  }
}

void Flow(float slow) {
  if (flow) {
    
    float dX,dY,dZ;

    translateX -= (dX = slow*translateX);
    translateY -= (dY = slow*translateY);
    translateZ -= (dZ = slow*translateZ);
    cam.Translate(dX,dY,dZ);

    slideX -= (dX = slow*slideX);
    slideY -= (dY = slow*slideY);
    slideZ -= (dZ = slow*slideZ);
    cam.Slide(dX,dY,dZ);

    rotateX -= (dX = slow*rotateX);
    rotateY -= (dY = slow*rotateY);
    rotateZ -= (dZ = slow*rotateZ);
    cam.Rotate(dX,X_AXIS);
    cam.Rotate(dY,Y_AXIS);
    cam.Rotate(dZ,Z_AXIS);

    cam.Translate(0,heightfactor*swingampl*sin(swing+=swingstep),0);
    glutPostRedisplay();
  }
}

void FixedParameters() {
  float mat_ambient[]   = { 0.3, 0.3, 0.3, 1.0 };
  float mat_diffuse[]   = { 0.7, 0.7, 0.7, 1.0 };
  float mat_specular[]  = { 0.0, 0.0, 0.0, 1.0 };
  float mat_shininess[] = { 0.0 };
  glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT , mat_ambient);
  glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE , mat_diffuse);
  glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, mat_specular);
  glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, mat_shininess);

  float light_ambient0[]  = { 1.0, 1.0, 1.0, 0.0 };
  float light_diffuse0[]  = { 1.0, 1.0, 1.0, 0.0 };
  float light_specular0[] = { 1.0, 0.0, 0.0, 0.0 };
  glLightfv(GL_LIGHT0, GL_AMBIENT, light_ambient0);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diffuse0);
  glLightfv(GL_LIGHT0, GL_SPECULAR,light_specular0);

  float light_ambient1[]  = { 1.0, 1.0, 1.0, 0.0 };
  float light_diffuse1[]  = { 1.0, 1.0, 1.0, 0.0 };
  float light_specular1[] = { 1.0, 1.0, 1.0, 0.0 };
  glLightfv(GL_LIGHT1, GL_AMBIENT, light_ambient1);
  glLightfv(GL_LIGHT1, GL_DIFFUSE, light_diffuse1);
  glLightfv(GL_LIGHT1, GL_SPECULAR,light_specular1);

  glLightf(GL_LIGHT1, GL_SPOT_CUTOFF, 50);
  glLightf(GL_LIGHT1, GL_SPOT_EXPONENT, 75);
  glLightf(GL_LIGHT1, GL_CONSTANT_ATTENUATION, 0);
  glLightf(GL_LIGHT1, GL_LINEAR_ATTENUATION, 0.02);
  glLightf(GL_LIGHT1, GL_QUADRATIC_ATTENUATION, 0);

  glDepthFunc(GL_LEQUAL);
  glEnable(GL_DEPTH_TEST);

  glCullFace(GL_FRONT);
  glEnable(GL_CULL_FACE);

  glHint(GL_POINT_SMOOTH_HINT,GL_FASTEST);
  glHint(GL_LINE_SMOOTH_HINT,GL_FASTEST);
  glHint(GL_POLYGON_SMOOTH_HINT,GL_FASTEST);
  glHint(GL_PERSPECTIVE_CORRECTION_HINT,GL_FASTEST);

  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
}

void VariableParameters() {
  if (solid) terrain = list_solid;
  else terrain = list_wire;
  if (textures && solid) glEnable(GL_TEXTURE_2D);
  else glDisable(GL_TEXTURE_2D);
  if (light) { glEnable(GL_LIGHT0); }
  else { glDisable(GL_LIGHT0); }
  if (reflector) { glEnable(GL_LIGHT1); }
  else { glDisable(GL_LIGHT1); }
  if (light || reflector) {  glEnable(GL_LIGHTING); }
  else { glDisable(GL_LIGHTING); }
  if (antialias) {
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
  }
  else {
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST);
  }
  if (smooth) glShadeModel(GL_SMOOTH);
  else glShadeModel(GL_FLAT);
  float light_position[4];
  switch (lightstyle) {
    case MORNING :   light_position[0] =  10.0;
                     light_position[1] = -10.0;
                     light_position[2] =   7.0;
                     break;
    case NOON :      light_position[0] =   0.0;
                     light_position[1] = -10.0;
                     light_position[2] =   8.0;
                     break;
    case AFTERNOON : light_position[0] = -10.0;
                     light_position[1] = -10.0;
                     light_position[2] =   7.0;
                     break;
  }
  light_position[3] = 0.0;
  glLightfv(GL_LIGHT0, GL_POSITION, light_position);
  if (reflector) {
    light_position[0] =  cam.eye[0];
    light_position[1] =  cam.eye[1];
    light_position[2] =  cam.eye[2];
    light_position[3] =  1.0;
    glLightfv(GL_LIGHT1, GL_POSITION, light_position);
    light_position[0] =  cam.look[0]-cam.eye[0];
    light_position[1] =  cam.look[1]-cam.eye[1];
    light_position[2] =  cam.look[2]-cam.eye[2];
    glLightfv(GL_LIGHT1, GL_SPOT_DIRECTION, light_position);
  }
  if (win_info) {
    glutSetWindow(win_info);
    if (status) glutShowWindow();
    else glutHideWindow();
    glutSetWindow(win_main);
  }
}

void myDisplay() {
  FixedParameters();
  VariableParameters();

  if (stereo) {
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    cam.Translate(-0.5*eyes,0,0);
    glColorMask(true,false,false,false);
    glCallList(terrain);
    cam.Translate(+eyes,0,0);
    glColorMask(false,false,true,false);
    glClear(GL_DEPTH_BUFFER_BIT);
    glCallList(terrain);
    cam.Translate(-0.5*eyes,0,0);
    glColorMask(true,true,true,true);
  }
  else {
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glCallList(terrain);
  }

  glCallList(list_track);

  glutSwapBuffers();
  if (status && win_info) {
    glutSetWindow(win_info);
    glutPostRedisplay();
    glutSetWindow(win_main);
  }

//  cam.DebugPrint();
}

void myReshape(int width, int height) {
  glViewport(0, 0, width, height);
  ratio = (float)width/(float)height;
  Transformations();
  if (win_info) {
    glutSetWindow(win_info);
    glutReshapeWindow(width,15);
    glutSetWindow(win_main);
  }
}

void subDisplay() {
  glutSetWindow(win_info);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  char tmp[256];
  glColor3f(1.0,1.0,1.0);

  sprintf(tmp,"Northing: %9.1f",cam.eye[1]/gridY*heights_stepY+heights_startY);
  DrawString(tmp,5,4);

  sprintf(tmp,"Easting: %9.1f",cam.eye[0]/gridY*heights_stepX+heights_startX);
  DrawString(tmp,130,4);

  sprintf(tmp,"Altitude above sea level: %6.1f",cam.eye[2]/heightfactor);
  DrawString(tmp,250,4);

  float h;
  if (InterpolateHeight(cam.eye[0]/gridX,cam.eye[1]/gridY,h)) {
    sprintf(tmp,"Altitude above terrain: %6.1f",cam.eye[2]/heightfactor-h);
    DrawString(tmp,445,4);
  }

  glutSwapBuffers();
  glutSetWindow(win_main);
}

void subReshape(int width, int height) {
  glViewport(0,0,width,height);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluOrtho2D(0,width,0,height);
}

void mySpecial(int key, int x, int y) {
  if (glutGetModifiers() & GLUT_ACTIVE_ALT)
    switch (key) {
      case GLUT_KEY_UP:        Slide(+speedXY*(cam.look[0]-cam.eye[0]),+speedXY*(cam.look[1]-cam.eye[1]),0); return;
      case GLUT_KEY_DOWN:      Slide(-speedXY*(cam.look[0]-cam.eye[0]),-speedXY*(cam.look[1]-cam.eye[1]),0); return;
      case GLUT_KEY_LEFT:      Slide(-speedXY*(cam.look[1]-cam.eye[1]),+speedXY*(cam.look[0]-cam.eye[0]),0); return;
      case GLUT_KEY_RIGHT:     Slide(+speedXY*(cam.look[1]-cam.eye[1]),-speedXY*(cam.look[0]-cam.eye[0]),0); return;
    }
  else
    switch (key) {
      case GLUT_KEY_UP:        Translate(0,0,-speedXY); return;
      case GLUT_KEY_DOWN:      Translate(0,0,+speedXY); return;
      case GLUT_KEY_LEFT:      Rotate(+rotate,Y_AXIS);  return;
      case GLUT_KEY_RIGHT:     Rotate(-rotate,Y_AXIS);  return;
      case GLUT_KEY_PAGE_DOWN: Rotate(+rotate,Z_AXIS);  return;
      case GLUT_KEY_HOME:      Rotate(+rotate,X_AXIS);  return;
      case GLUT_KEY_END:       Rotate(-rotate,X_AXIS);  return;
    }
}

void myKeyboard(unsigned char key, int x, int y) {
  if (glutGetModifiers() & GLUT_ACTIVE_ALT)
    switch (key) {
      case 'a': case 'A': Slide(0,0,+speedZ); return;
      case 'z': case 'Z': Slide(0,0,-speedZ); return;
    }
  else 
    switch (key) {
      case '+':           angle+=1; Transformations(); break;
      case '-':           angle-=1; Transformations(); break;
      case '[':           eyes -= 0.01; break;
      case ']':           eyes += 0.01; break;
      case 'q': case 'Q': exit();
      case 127:           Rotate(-rotate,Z_AXIS);  return;
      case 'a': case 'A': Translate(0,+speedZ,0);  return;
      case 'z': case 'Z': Translate(0,-speedZ,0);  return;
      case ',': case '<': Translate(-speedXY,0,0); return;
      case '.': case '>': Translate(+speedXY,0,0); return;
      case 's': case 'S': solid = !solid; break;
      case 'w': case 'W': stereo = !stereo; break;
      case 'r': case 'R': reflector = !reflector; break;
    }
  glutPostRedisplay();
}

void myMenu(int item) {
  switch(item) {
    case  1: solid = true; break;
    case  2: solid = false; break;
    case  3: textures = true; antialias = true; break;
    case  4: textures = true; antialias = false; break;
    case  5: textures = false; break;
    case  6: light = true; lightstyle = MORNING; break;
    case  7: light = true; lightstyle = NOON; break;
    case  8: light = true; lightstyle = AFTERNOON; break;
    case  9: light = false; break;
    case 10: smooth = false; break;
    case 11: smooth = true; break;
    case 12: flow = false; return;
    case 13: flow = true; return;
    case 14: status = true; break;
    case 15: status = false; break;
    case 16: stereo = true; break;
    case 17: stereo = false; break;
    case 18: reflector = true; break;
    case 19: reflector = false; break;
    case 20: exit();
  }
  glutPostRedisplay();
}

void myMouse(int button, int state, int x, int y) {
  switch(button) {
    case GLUT_LEFT_BUTTON:
      break;
    case GLUT_MIDDLE_BUTTON:
      break;
    case GLUT_RIGHT_BUTTON:
      break;
  }
  switch(state) {
    case GLUT_UP:
      break;
    case GLUT_DOWN:
      break;
  }
}

void myPassiveMotion(int x, int y) {
}

void myTimer(int millis) {
  Flow(0.33);
  glutTimerFunc(millis,myTimer,millis);
}

void myIdle() {
  timeb actual;
  ftime(&actual);
  long diff;
  if (last.time) {
    diff = 1000*(actual.time-last.time)+(actual.millitm-last.millitm);
  }
  else {
    diff = swingdelay;
  }
  Flow(1-powf(0.66,diff/swingdelay));
  last = actual;
}

void SetupData() {
  heights_startX = heights_stepX*(surface_startX/heights_stepX);
  heights_stopX  = heights_stepX*(surface_stopX/heights_stepX);
  heights_startY = heights_stepY*(surface_startY/heights_stepY);
  heights_stopY  = heights_stepY*(surface_stopY/heights_stepY);

  heightsX = (heights_stopX-heights_startX)/heights_stepX;
  heightsY = (heights_stopY-heights_startY)/heights_stepY;

  heightfactor = plasticity*gridX/heights_stepX;
  
  if (surface) {
    overlapX = 1*pixelsX/heightsX+1;
    overlapY = 1*pixelsY/heightsY+1;
  }

  if (startX==-1) {
    startX = 0;        startY = 0;
    stopX  = heightsX; stopY  = heightsY;
  }
}

int main(int argc, char **argv) {

// BASIC INITIALIZATION
  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH);
  if (1) {
    glutInitWindowPosition(5, 5);
    glutInitWindowSize(440, 330);
    win_main = glutCreateWindow("3D DEM Viewer by Oto Valek");
    win_info = glutCreateSubWindow(win_main, 0, 0, 100, 15);
    glutDisplayFunc(subDisplay);
    glutReshapeFunc(subReshape);
    glutSetWindow(win_main);
  }
  else {
//    glutGameModeString("640x480:16@75");
//    glutEnterGameMode();
    glutSetCursor(GLUT_CURSOR_NONE);
  }

// CALLBACKS
  glutDisplayFunc(myDisplay);
  glutReshapeFunc(myReshape);
  glutSpecialFunc(mySpecial);
  glutKeyboardFunc(myKeyboard);
  glutMouseFunc(myMouse);
  glutPassiveMotionFunc(myPassiveMotion);
//  glutTimerFunc(0,myTimer,(int)swingdelay);
  glutIdleFunc(myIdle);

// MENU
  int menu_surface = glutCreateMenu(myMenu);
  glutAddMenuEntry("Solid", 1);
  glutAddMenuEntry("Wire", 2);
  int menu_textures = glutCreateMenu(myMenu);
  glutAddMenuEntry("On - antialiasing", 3);
  glutAddMenuEntry("On - no antialiasing", 4);
  glutAddMenuEntry("Off", 5);
  int menu_lighting = glutCreateMenu(myMenu);
  glutAddMenuEntry("On - morning", 6);
  glutAddMenuEntry("On - noon", 7);
  glutAddMenuEntry("On - afternoon ", 8);
  glutAddMenuEntry("Off", 9);
  int menu_shade = glutCreateMenu(myMenu);
  glutAddMenuEntry("Flat", 10);
  glutAddMenuEntry("Smooth", 11);
  int menu_movement = glutCreateMenu(myMenu);
  glutAddMenuEntry("Quick", 12);
  glutAddMenuEntry("Smooth", 13);
  int menu_status = glutCreateMenu(myMenu);
  glutAddMenuEntry("On", 14);
  glutAddMenuEntry("Off", 15);
  int menu_stereo = glutCreateMenu(myMenu);
  glutAddMenuEntry("On", 16);
  glutAddMenuEntry("Off", 17);
  int menu_reflector = glutCreateMenu(myMenu);
  glutAddMenuEntry("On", 18);
  glutAddMenuEntry("Off", 19);
  glutCreateMenu(myMenu);
  glutAddSubMenu("Surface type", menu_surface);
  glutAddSubMenu("Textures", menu_textures);
  glutAddSubMenu("Lighting", menu_lighting);
  glutAddSubMenu("Shade model", menu_shade);
  glutAddSubMenu("Movement", menu_movement);
  glutAddSubMenu("Status", menu_status);
  glutAddSubMenu("Stereo", menu_stereo);
  glutAddSubMenu("Reflector", menu_reflector);
  glutAddMenuEntry("Exit", 20);
  glutAttachMenu(GLUT_RIGHT_BUTTON);

// RESET DEFINITION
  surface_stopY = heights_stepX = startX = pixelsX = -1;
  name_heights[0] = name_surface[0] = name_palette[0] = name_track[0] = 0;

// LOAD DEFINITION
  printf("Loading def file ..\n");
  FILE *def = fopen(argv[1],"rb");
  if (!def) {
    printf("Error opening '%s'\n",argv[1]); exit(1);
  }
  while (!feof(def)) {
    char line[256];
    fgets(line,256,def);
    int cnt = 0;
    cnt += sscanf(line,"DEMRANGE %d %d %d %d",&surface_stopY,&surface_startX,&surface_startY,&surface_stopX);
    cnt += sscanf(line,"DEMSTEP %d %d",&heights_stepX,&heights_stepY);
    cnt += sscanf(line,"DEMFILE %s",name_heights);
    cnt += sscanf(line,"DEMLIMIT %d %d %d %d",&startX,&startY,&stopX,&stopY);
    cnt += sscanf(line,"TEXSIZE %d %d",&pixelsX,&pixelsY);
    cnt += sscanf(line,"TEXFILE %s",name_surface);
    cnt += sscanf(line,"TEXPAL %s",name_palette);
    cnt += sscanf(line,"TRACK %s",name_track);
    if (cnt==0) printf("Unrecognized line : %s",line);
  }
  fclose(def);
  if (!(surface_stopY && heights_stepX && name_heights)) {
    printf("Incomplete definition\n"); exit(1);
  }
  
  surface = (pixelsX && pixelsY && name_surface[0]);
  SetupData();

// ALLOCATE DATA & LOAD FILES
  printf("Loading heights model .. %d x %d\n",heightsX,heightsY);
  heights = (unsigned short*)LoadWholeFile(name_heights);
  normals = new Vector3f[heightsX*heightsY];
  if (name_palette[0]) {
    printf("Loading palette ..\n");
    palette = (unsigned char*)LoadWholeFile(name_palette);
  }
  if (name_track[0]) {
    printf("Loading track ..\n");
    track = LoadTrack(name_track);
  }

// INITIAL POSITION
  const float AZIMUTH = 45;
  const float SLOPE   = 15;
  Vector3f corner;
  GetTerrainPoint(startX,startY,corner);
  cam.eye[0]  = corner[0];
  cam.eye[1]  = corner[1];
  cam.eye[2]  = corner[2]+500.0*heightfactor;
  cam.up[0]   = cam.eye[0] + sin(SLOPE*DEG2RAD)*cos(AZIMUTH*DEG2RAD);
  cam.up[1]   = cam.eye[1] + sin(SLOPE*DEG2RAD)*sin(AZIMUTH*DEG2RAD);
  cam.up[2]   = cam.eye[2] + cos(SLOPE*DEG2RAD);
  cam.side[0] = cam.eye[0] + sin(AZIMUTH*DEG2RAD);
  cam.side[1] = cam.eye[1] - cos(AZIMUTH*DEG2RAD);
  cam.side[2] = cam.eye[2] + 0.0;
  cam.look[0] = cam.eye[0] + cos(SLOPE*DEG2RAD)*cos(AZIMUTH*DEG2RAD);
  cam.look[1] = cam.eye[1] + cos(SLOPE*DEG2RAD)*sin(AZIMUTH*DEG2RAD);
  cam.look[2] = cam.eye[2] - sin(SLOPE*DEG2RAD);

// GENERATE MODEL
  FixedParameters();
  GenerateTerrain();
  GenerateTrack();

// DELETE ALLOCATED DATA
  delete palette;

// GLUT MAIN EVENT LOOP
  glutMainLoop();

  return 0;
}

void exit() {

// DELETE ALLOCATED DATA
  delete normals;
  delete heights;

  exit(0);
}
